#!/usr/bin/python3.10
# -*-coding:Utf-8 -*

#==============================================================================
# TENET: SHACL Inference Engine
#------------------------------------------------------------------------------
# Module to execute the extraction process, 
# by applying the transduction schemes with the SHACL inference engine
#==============================================================================

#==============================================================================
# Importing required modules
#==============================================================================

from subprocess import Popen, PIPE
import os
import time, datetime


#==============================================================================
# Parameters
#==============================================================================

# None


#==============================================================================
# Inference Run
#==============================================================================

def run_command(cmd):
    with Popen(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True) as p:
        p.poll()
        p.stdout.flush()
        p.stderr.flush()
        stdout, stderr = p.communicate()
    return p.returncode, stdout, stderr


def define_shacl_command(script_name, ttl_file):
    """ Define command to run SHACL inference, by adapting the file names 
        to the OS. """

    cmd = ['sh', script_name, '-datafile', ttl_file]
    
    if os.name != 'posix': # OS is considering to be Windows
        script_name = script_name.replace('/', '\\')
        script_name = script_name.replace('.sh', '.bat')
        ttl_file = ttl_file.replace('/', '\\')
        cmd = [script_name, '-datafile', ttl_file]
        
    return cmd


def shaclInfer(config, ttl_file, mode, ttl_rules_file = ''):
    """ Run SHACL infer sh script. 
        mode argument can take the values 'infer' or 'validate' """
    
    shacl_bin_dir = config.shacl_bin_dir
    script_basename = 'shacl' + mode + '.sh'
    script_name = shacl_bin_dir + '/' + script_basename
    # -- old --- cmd = ['sh', script_name, '-datafile', ttl_file]
    cmd = define_shacl_command(script_name, ttl_file)
    if ttl_rules_file != '':
        cmd += ['-shapesfile', ttl_rules_file]
        
    print("--- Run command: " + str(cmd))
    code, out, err = run_command(cmd)
    if code != 0:
         print("Error in SHACL script: \n\n" + err)
    else:
        return(out)
     
    
    
#==============================================================================
# Inference Step
#==============================================================================
    
clearExecutionInstances = """
    PREFIX cts: <https://unsel.tetras-libre.fr/tenet/transduction-schemes#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

    DELETE {?x rdf:type ?c}
    WHERE {
        ?c rdfs:subClassOf* cts:Transduction_Schemes .
        ?x rdf:type ?c .
    }
"""

addExecutionInstance = """
    PREFIX cts: <https://unsel.tetras-libre.fr/tenet/transduction-schemes#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

    INSERT DATA {{<exec_instance> rdf:type {}}}
"""
    
    
def apply_step(config, graph, step):
    try:
        print("\n" + "- Apply inference step: " + step)
    
        print("-- Preparation")
        step = 'shacl_' + step
        step_ref = "cts:" + step
        uuid_str = config.uuid_str
        output_dir = config.output_dir
        work_file = output_dir + uuid_str + '-' + step + ".ttl"
        base_ref = "http://" + uuid_str + '/' + step
    
        print("-- Initialization")
        graph.update(clearExecutionInstances)
        graph.update(addExecutionInstance.format(step_ref)) # ex. : step = 'cts:generation'
        graph.serialize(destination=work_file, base=base_ref, format='turtle')
    
        print("-- Apply SHACL inference in work graph")
        inferResult = shaclInfer(config, work_file, 'infer')
    
        print("-- Update dest graph with inference")
        graph.parse(data=inferResult)
        graph.serialize(destination=work_file, base=base_ref, format='turtle')
    
        return graph, inferResult
    
    except:
        print("!!! An exception occurred !!!")   
   
     
#==============================================================================
# Main Function
#==============================================================================   

def apply(config, work_graph):
    """ apply inference processes on the working graph """
    
    try:
        print("\n" + "[Tenet] SHACL extraction engine")
        exec_start = time.perf_counter()
        process_start = time.process_time()
    
        # -- Extraction
        work_graph, _ = apply_step(config, work_graph, 'preprocessing')
        work_graph, _ = apply_step(config, work_graph, 'net_extension')
        work_graph, inf_result = apply_step(config, work_graph, 'generation')
    
        # -- Write result 
        output_dir = config.output_dir
        uuid_str = config.uuid_str
        factoidPath = output_dir + uuid_str + '-shacl_factoid.ttl'
        with open(factoidPath, 'w') as outfile:
            outfile.write(inf_result)
            outfile.close()
        
        # -- Ending
        exec_end = time.perf_counter()
        process_end = time.process_time()
        exec_time = exec_end - exec_start
        exec_time_date = datetime.timedelta(seconds=exec_time)
        process_time = process_end - process_start
        process_time_date = datetime.timedelta(seconds=process_time)
        print("\n" + "- Ending SHACL inference application ")
        print("-- Total Execution Time = " + str(exec_time_date))
        print("-- Process Time = " + str(process_time_date))
    
    except:
        print("!!! An exception occurred !!!")     
    

