#!/usr/bin/python3.10
# -*-coding:Utf-8 -*

#==============================================================================
# TENET: main
#------------------------------------------------------------------------------
# Module providing the main methods of the TENET library.
#==============================================================================

import sys, os, glob
import shutil
from rdflib import Graph
import logging.config

# Set main file directory as reference
LIB_PATH = os.path.dirname(os.path.abspath(__file__)) + '/'
print('Running in ' + LIB_PATH)
os.chdir(LIB_PATH)
sys.path.insert(1, LIB_PATH)

from lib import config, structure
from lib import shacl_extraction, tenet_extraction
from utility.timer import timed

# -- Logging
logging.config.fileConfig('logging.conf', disable_existing_loggers=False)
logger = logging.getLogger('root')

# Configuration
CONFIG_FILE = "config.xml"
    
   
#==============================================================================
# Steps
#==============================================================================

def set_config(args, technical_dir_path):
    
    logger.info("-- Process Setting ")
    logger.info("----- Corpus source: {0} ({1})".format(args['source_corpus'], 
                                                        args['source_type']))
    logger.info("----- Base output dir: {0}".format(args['base_output_dir']))
    logger.info(f'----- technical dir path: {technical_dir_path}')
    logger.info("----- Ontology target (id): {0}".format(args['target_id']))
    logger.debug("----- Current path: {0}".format(os.getcwd()))
    logger.debug("----- Config file: {0}".format(CONFIG_FILE))
    
    process_config = config.Config(CONFIG_FILE, 
                           args['target_id'], 
                           args['source_corpus'], 
                           #target_ontology,
                           base_output_dir = args['base_output_dir'],
                           technical_dir_path = technical_dir_path
                           )
    process_config.source_type = args['source_type']
    # config.output_ontology_namespace = target_ontology_namespace
    
    process_config.engine = args['engine']
    
    logger.debug(process_config.get_full_config())
    
    return process_config


def init_process(config):
    
    # logger.info("-- Creating output target directory: " + config.output_dir)
    # os.makedirs(config.output_dir, exist_ok=True)
    
    logger.debug("-- Counting number of graph files (sentences) ")
    sentence_count = 0
    for file_ref in glob.glob(config.source_sentence_file, recursive = True):
        sentence_count += 1   
    logger.debug("----- Graph count: {0}".format(sentence_count))


def apply_extraction(config, sentence_file):
    os.makedirs(config.sentence_output_dir, exist_ok=True)
    work_graph = structure.prepare_sentence_work(config, sentence_file)
    _, new_triple_list = tenet_extraction.apply(config, work_graph)
    return new_triple_list
    
    
def generate_final_ontology(result_triple_list): 
    config.sentence_output_dir = ''
    logger.info("-- Making complete factoid graph by merging the result factoids")
    factoid_graph = Graph()
    for new_triple in result_triple_list:
        factoid_graph.add(new_triple) 
    logger.info(f'----- Total factoid number: {len(result_triple_list)}')
    return factoid_graph
   
    
def serialize_factoid_graph(config, factoid_graph, out_file_path=None):
    
    logger.info('-- Serializing graph to factoid string')
    base_ref = f'http://{config.uuid_str}/factoid'
    logger.info(f'----- Graph base: {base_ref}')
    ontology_turtle_string = factoid_graph.serialize(base=base_ref, format='turtle')
    
    if out_file_path is not None:
        logger.info('-- Serializing graph to factoid file')
        logger.info(f'----- Ontology Turtle File: {out_file_path}')
        factoid_graph.serialize(destination=out_file_path, base=base_ref, format='turtle')
        
    return ontology_turtle_string    
    
  
# -- old ----    
# @timed    
# def run_tenet_extraction(config):
     
#     logger.debug("-- Process level: sentence")  
            
#     sentence_dir = config.source_sentence_file
#     sentence_count = 0
#     result_triple_list = []
#     for sentence_file in glob.glob(sentence_dir, recursive = True):
#         sentence_count += 1
#         config.sentence_output_dir = '-' + str(sentence_count)
#         logger.info("     *** sentence {0} *** ".format(sentence_count))
#         os.makedirs(config.sentence_output_dir, exist_ok=True)
#         work_graph = structure.prepare_sentence_work(config, sentence_file)
#         # New extraction engine running
#         _, new_triple_list = tenet_extraction.apply(config, work_graph)
#         result_triple_list.extend(new_triple_list)
     
#     logger.info(' === Final Ontology Generation  === ')   
#     config.sentence_output_dir = ''
#     logger.info("-- Making complete factoid graph by merging sentence factoid graphs")
#     factoid_graph = Graph()
#     for new_triple in result_triple_list:
#         factoid_graph.add(new_triple) 
#     logger.info("----- Total factoid number: " + str(len(new_triple_list)))    
#     uuid_str = config.uuid_str
#     base_ref = "http://" + uuid_str + '/' + 'factoid'
#     logger.info("----- Graph base: {0}".format(base_ref)) 
#     factoid_file = config.output_file.replace('.ttl', '_factoid.ttl')
#     logger.info("-- Serializing graph to factoid file ({0})".format(factoid_file))
#     factoid_graph.serialize(destination=factoid_file, 
#                             base=base_ref, 
#                             format='turtle')
    

#==============================================================================
# AMR Main Methods 
#==============================================================================

@timed
def create_ontology_from_amrld_file(amrld_file_path, 
                                    base_ontology_path=None, 
                                    onto_prefix=None, 
                                    out_file_path=None, 
                                    technical_dir_path=None):
    """
    Method to create an ontology (as Turtle String) from a transduction 
    analysis of an AMRLD file.

    Parameters
    ----------
    amrld_file_path: a path to an AMR-LD Turtle File.
    base_ontology_path: a path to a Base Ontology Turtle File if defined.
    onto_prefix: the target ontology prefix if defined (if not defined a prefix based on the amrld filename is used).
    out_file_path: a file path where the output ontology is written if defined (the function still outputs the string). 
    technical_dir_path: a dir path where some technical and log files are written if defined.

    Returns
    -------
    Ontology Turtle String.

    """
    
    logger.info('[TENET] Extraction Processing')    
    
    # -- Process Initialization
    logger.info(' === Process Initialization === ')
    logger.info('-- current dir: {0}'.format(os.getcwd()))
    if onto_prefix is None: onto_prefix = 'DefaultTargetId'
    args = {
        'source_type': 'amr',
        'source_corpus': amrld_file_path,
        'target_id': onto_prefix,
        'base_output_dir': out_file_path,
        'engine': 'tenet'}
    config = set_config(args, technical_dir_path)
    init_process(config)
    
    # -- Extraction Processing using TENET Engine    
    logger.info(' === Extraction Processing using New TENET Engine === ')
    config.sentence_output_dir = f'-0'
    result_triple_list = apply_extraction(config, amrld_file_path)
    
    # -- Final Ontology Generation (factoid_graph)
    logger.info(' === Final Ontology Generation  === ') 
    factoid_graph = generate_final_ontology(result_triple_list)
    ontology_turtle_string = serialize_factoid_graph(config, factoid_graph, out_file_path)
        
    # -- Done
    logger.info(' === Done === ')
    if config.technical_dir_path is not None:
        log_file_name = 'tenet.log'
        dest_file_path = f'{config.technical_dir_path}{log_file_name}'
        shutil.copy(log_file_name, dest_file_path)
    
    return ontology_turtle_string


@timed
def create_ontology_from_amrld_dir(amrld_dir_path, 
                                   base_ontology_path=None, 
                                   onto_prefix=None, 
                                   out_file_path=None, 
                                   technical_dir_path=None):
    """
    Method to create an ontology (as Turtle String) from a transduction 
    analysis of an AMRLD file.

    Parameters
    ----------
    amrld_file_path: a path to an AMR-LD Turtle File.
    base_ontology_path: a path to a Base Ontology Turtle File if defined.
    onto_prefix: the target ontology prefix if defined (if not defined a prefix based on the amrld filename is used).
    out_file_path: a file path where the output ontology is written if defined (the function still outputs the string). 
    technical_dir_path: a dir path where some technical and log files are written if defined.

    Returns
    -------
    Dictionary [filename -> Ontology Turtle String].
    Complete Ontology Turtle String (synthesis of all ontology)

    """

    logger.info('[TENET] Extraction Processing')    
    
    # -- Process Initialization
    logger.info(' === Process Initialization === ')
    logger.info('-- current dir: {0}'.format(os.getcwd()))
    if onto_prefix is None: onto_prefix = 'DefaultTargetId'
    args = {
        'source_type': 'amr',
        'source_corpus': amrld_dir_path,
        'target_id': onto_prefix,
        'base_output_dir': out_file_path,
        'engine': 'tenet'}
    config = set_config(args, technical_dir_path)
    init_process(config)
    
    # -- Extraction Processing using TENET Engine    
    logger.info(' === Extraction Processing using New TENET Engine === ')
    sentence_dir = config.source_sentence_file
    sentence_count = 0
    result_triple_list = []
    for sentence_file in glob.glob(sentence_dir, recursive = True):
        sentence_count += 1
        logger.info(f'     *** sentence {sentence_count} *** ')
        config.sentence_output_dir = f'-{sentence_count}'
        new_triple_list = apply_extraction(config, sentence_file)
        result_triple_list.extend(new_triple_list)
    
    # -- Final Ontology Generation (factoid_graph)
    logger.info(' === Final Ontology Generation  === ') 
    factoid_graph = generate_final_ontology(result_triple_list)
    ontology_turtle_string = serialize_factoid_graph(config, factoid_graph, out_file_path)
        
    # -- Done
    logger.info(' === Done === ')
    if config.technical_dir_path is not None:
        log_file_name = 'tenet.log'
        dest_file_path = f'{config.technical_dir_path}{log_file_name}'
        shutil.copy(log_file_name, dest_file_path)
    
    return ontology_turtle_string
    

#==============================================================================
# UNL Main Methods 
#==============================================================================

def create_ontology_from_unlrdf_file(unlrdf_file_path,
                                     base_ontology_path,
                                     ontology_seed_path,
                                     onto_prefix=None, 
                                     out_file_path=None, 
                                     technical_dir_path=None):
    """
    Method to create an ontology (as Turtle String) from a transduction 
    analysis of an UNL-RDF file, using SHACL engine.

    Parameters
    ----------
    amrld_file_path: a path to an AMR-LD Turtle File.
    base_ontology_path: a path to a Base Ontology Turtle File if defined.
    onto_prefix: the target ontology prefix if defined (if not defined a prefix based on the amrld filename is used).
    out_file_path: a file path where the output ontology is written if defined (the function still outputs the string). 
    technical_dir_path: a dir path where some technical and log files are written if defined.

    Returns
    -------
    Ontology Turtle String.

    """
    
    logger.info('[TENET] Extraction Processing')    
    
    # -- Process Initialization
    logger.info(' === Process Initialization === ')
    logger.info('-- current dir: {0}'.format(os.getcwd()))
    if onto_prefix is None: onto_prefix = 'DefaultTargetId'
    args = {
        'source_type': 'unl',
        'source_corpus': unlrdf_file_path,
        'target_id': onto_prefix,
        'base_output_dir': out_file_path,
        'engine': 'shacl'}
    config = set_config(args)
    config.process_level = 'document'
    config.frame_ontology_file = base_ontology_path
    config.frame_ontology_seed_file = ontology_seed_path
    #config.output_ontology_namespace = onto_prefix
    init_process(config)
    
    # -- Extraction Processing using SHACL Engine  
    logger.info(' === Extraction Processing using SHACL Engine === ')
    logger.debug("-- Process level: document")    
    work_graph = structure.prepare_document_work(config)
    shacl_extraction.apply(config, work_graph)
    
    # -- Done
    logger.info(' === Done === ')
    dest_dir = config.output_dir
    filePath = shutil.copy('tenet.log', dest_dir)