#!/usr/bin/python3.10
# -*-coding:Utf-8 -*

#==============================================================================
# TENET: AMR CTR at 'Net Expansion' level
#------------------------------------------------------------------------------
# Module grouping compositional transduction rule_sets (CTR) for the analysis of 
# AMR structures, at 'Net Expansion' level
#==============================================================================

#==============================================================================
# CTR Set (CTR as Sparql querries)
#==============================================================================

rule_set = {}


# ---------------------------------------------
# Net Expansion: Composite Class Net from Phenomena Net
# ---------------------------------------------

rule_set['create-composite-class-net-from-phenomena-1'] = {
    'label': "create-composite-class-net-from-phenomena-1",
    'comment': ("Create Composite Class Net from relations: ", 
                " (1) between classes and DEGREE phenomena",
                " (2) targetting attribute (:ARG2) and degree (:ARG3)"),
    'construction': """
        # -- New Class Net (on attributeNet base leaf)
        ?newNet a net:Composite_Class_Net ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:coverArgNode ?degreeLeaf ;
            net:hasClassName ?newClassName ;
            net:hasMotherClassNet ?attributeNet.
        ?degreePhenomenaNet net:targetSubAttribute ?newNet.
        
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify degree(attribute, degree)
        ?degreePhenomenaNet a net:Phenomena_Net ;
            net:coverBaseNode ?phenomenaLeaf ;
            net:targetAttribute ?attributeNet ;
            net:targetDegree ?degreeNet ;
            net:hasPhenomenaType amr:phenomena_degree.
        ?attributeNet a [rdfs:subClassOf* net:Class_Net] ;
        		net:coverBaseNode ?baseLeaf ;
    			net:hasClassName ?attributeName.
        ?degreeNet a [rdfs:subClassOf* net:Class_Net] ;
        		net:coverBaseNode ?degreeLeaf ;
    			net:hasClassName ?degreeName.
                
        # -- Identify structure
        ?root a amr:AMR_Root ;
            amr:hasSentenceID ?sentenceRef.
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
		# -- New Class Name
        BIND (CONCAT(str(?degreeName), '-', 
               str(?attributeName)) AS ?newClassName).
       	
        # -- New Class Net
		BIND (CONCAT(str(net:), 'compositeClass') AS ?n1).
		BIND (CONCAT(?n1, '_', ?newClassName) AS ?n2).
        BIND (uri(?n2) AS ?newNet).
    """
}

rule_set['create-composite-class-net-from-phenomena-2'] = {
    'label': "create-composite-class-net-from-phenomena-2",
    'comment': ("Create Composite Class Net from relations: ", 
                " (1) between classes and DEGREE phenomena",
                " (2) targetting domain (:ARG1) and attribute (:ARG2)"),
    'construction': """
        # -- New Class Net (on domainNet base leaf)
        ?newNet a net:Composite_Class_Net ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:coverArgNode ?attributeLeaf ;
            net:hasClassName ?newClassName ;
            net:hasMotherClassNet ?domainNet.
        ?degreePhenomenaNet net:targetSubDomain ?newNet.
        
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify degree(domain, attribute)
        ?degreePhenomenaNet a net:Phenomena_Net ;
            net:coverBaseNode ?phenomenaLeaf ;
            net:targetDomain ?domainNet ;
            net:targetAttribute ?attributeNet ;
            net:hasPhenomenaType amr:phenomena_degree.
        ?domainNet a [rdfs:subClassOf* net:Class_Net] ;
            	net:coverBaseNode ?baseLeaf ;
            net:hasClassName ?domainName.
        ?attributeNet a [rdfs:subClassOf* net:Class_Net] ;
            	net:coverNode ?attributeLeaf ;
    			net:hasClassName ?attributeName.
                
        # -- Identify structure
        ?root a amr:AMR_Root ;
            amr:hasSentenceID ?sentenceRef.
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
		# -- New Class Name
        BIND (CONCAT(str(?attributeName), '-', 
               str(?domainName)) AS ?newClassName).
       	
        # -- New Class Net
		BIND (CONCAT(str(net:), 'compositeClass') AS ?n1).
		BIND (CONCAT(?n1, '_', ?newClassName) AS ?n2).
        BIND (uri(?n2) AS ?newNet).
    """
}

rule_set['create-composite-class-net-from-phenomena-3'] = {
    'label': "create-composite-class-net-from-phenomena-3",
    'comment': ("Create Composite Class Net from relations: ", 
                " (1) between classes and DEGREE phenomena",
                " (2) targetting sub-domain and sub-attribute"),
    'construction': """
        # -- New Class Net (on domainNet base leaf)
        ?newNet a net:Composite_Class_Net ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:coverArgNode ?subAttributeLeaf ;
            net:hasClassName ?newClassName ;
            net:hasMotherClassNet ?subDomainNet.
        ?degreePhenomenaNet net:targetSubDomain ?newNet.
        
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify degree(sub-domain, sub-attribute)
        ?degreePhenomenaNet a net:Phenomena_Net ;
            net:coverBaseNode ?phenomenaLeaf ;
            net:targetDomain ?domainNet ;
            net:targetSubDomain ?subDomainNet ;
            net:targetSubAttribute ?subAttributeNet ;
            net:hasPhenomenaType amr:phenomena_degree.
        ?subDomainNet a [rdfs:subClassOf* net:Class_Net] ;
            	net:coverBaseNode ?baseLeaf ;
            net:hasClassName ?subDomainName.
        ?subAttributeNet a [rdfs:subClassOf* net:Class_Net] ;
            net:coverNode ?subAttributeLeaf ;
    			net:hasClassName ?subAttributeName.
                
        # -- Identify structure
        ?root a amr:AMR_Root ;
            amr:hasSentenceID ?sentenceRef.
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
		# -- New Class Name
        BIND (CONCAT(str(?subAttributeName), '-', 
               str(?subDomainName)) AS ?newClassName).
       	
        # -- New Class Net
		BIND (CONCAT(str(net:), 'compositeClass') AS ?n1).
		BIND (CONCAT(?n1, '_', ?newClassName) AS ?n2).
        BIND (uri(?n2) AS ?newNet).
    """
}

rule_set['create-composite-class-net-from-phenomena-4'] = {
    'label': "create-composite-class-net-from-phenomena-4",
    'comment': ("Create Composite Class Net from relations: ", 
                " (1) between classes and DEGREE phenomena",
                " (2) targetting sub-domain and superlative"),
    'construction': """
        # -- New Class Net (on subDomainNet base leaf)
        ?newNet a net:Composite_Class_Net ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?superlativeLeaf ;
            net:coverArgNode ?baseLeaf ;
            net:hasClassName ?newClassName ;
            net:hasMotherClassNet ?superlativeNet.
        ?degreePhenomenaNet net:targetSubDomain ?newNet.
        
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify degree(sub-domain, superlative)
        ?degreePhenomenaNet a net:Phenomena_Net ;
            net:coverBaseNode ?phenomenaLeaf ;
            net:targetSubDomain ?subDomainNet ;
            net:targetSuperlative ?superlativeNet ;
            net:hasPhenomenaType amr:phenomena_degree.
        ?subDomainNet a [rdfs:subClassOf* net:Class_Net] ;
            	net:coverBaseNode ?baseLeaf ;				
            net:coverNodeCount ?subDomainPrecision ;
            net:hasClassName ?subDomainName.
        FILTER NOT EXISTS { 
            ?otherSubDomainNet a [rdfs:subClassOf* net:Class_Net] ;
        			net:coverBaseNode ?baseLeaf ;									
				net:coverNodeCount ?otherSubDomainPrecision.											
			FILTER (?otherSubDomainPrecision > ?subDomainPrecision). 
            }	
        ?superlativeNet a [rdfs:subClassOf* net:Class_Net] ;
            	net:coverBaseNode ?superlativeLeaf ;						
            net:coverNodeCount ?superlativePrecision ;
            net:hasClassName ?superlativeName.
        FILTER NOT EXISTS { 
            ?otherSuperlativeNet a [rdfs:subClassOf* net:Class_Net] ;
        			net:coverBaseNode ?superlativeLeaf ;									
				net:coverNodeCount ?otherSuperlativePrecision.											
			FILTER (?otherSuperlativePrecision > ?superlativePrecision). 
            }		

        # -- Identify structure									
        ?root a amr:AMR_Root ;
            amr:hasSentenceID ?sentenceRef.
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
		# -- New Class Name
        BIND (CONCAT(str(?subDomainName), '-in-', 
               str(?superlativeName)) AS ?newClassName).
       	
        # -- New Class Net
		BIND (CONCAT(str(net:), 'compositeClass') AS ?n1).
		BIND (CONCAT(?n1, '_', ?newClassName) AS ?n2).
        BIND (uri(?n2) AS ?newNet).
    """
}