#!/usr/bin/python3.10
# -*-coding:Utf-8 -*

#==============================================================================
# TENET: AMR CTR at 'Net Expansion' level
#------------------------------------------------------------------------------
# Module grouping compositional transduction rule_sets (CTR) for the analysis of 
# AMR structures, at 'Net Expansion' level
#==============================================================================

#==============================================================================
# CTR Set (CTR as Sparql querries)
#==============================================================================

rule_set = {}

 
# ---------------------------------------------
# Net Expansion: Composite Property Net from Property Net
# ---------------------------------------------

rule_set['create-composite-property-net-from-property-1'] = {
    'label': "create-composite-class-net-from-property-1",
    'comment': ("Create Composite Property Net from relations: ", 
                " (1) between two core properties",
                " (2) with :ARG1 relation"),
    'construction': """
        # -- New Property Net
        ?newNet a net:Composite_Property_Net ;
            net:isCoreRoleLinked true ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:coverArgNode ?property2Leaf ;
            net:hasPropertyType ?property1Type ;
            net:hasPropertyName ?newPropertyName ;
            net:hasMotherClassNet ?property2Net.
            #net:hasRestriction ?restrictionNet.
            
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify ARG1(p1:Property, p2:Property)
        ?property1Net a [rdfs:subClassOf* net:Property_Net] ;
            net:isCoreRoleLinked true ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:hasPropertyType ?property1Type ;
            net:hasPropertyName ?property1Name.
        ?baseLeaf ?r1 ?property2Leaf.
        ?r1 amr:hasAmrRole amr:role_ARG1.
        ?property2Net a [rdfs:subClassOf* net:Property_Net] ;
            net:isCoreRoleLinked true ;
			net:coverBaseNode ?property2Leaf ;
			net:hasPropertyName ?property2Name.
            
        # -- condition: disjoint cover
        FILTER NOT EXISTS {?property1Leaf net:coverNode ?node.
			               ?property2Leaf net:coverNode ?node.}
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
        	# -- New Property Name
        BIND (CONCAT(?property1Name, '-', ?property2Name) AS ?newPropertyName).
        
        	# -- New Property Net
        BIND (CONCAT(str(net:), 'compositeProperty') AS ?n1).
        BIND (CONCAT(?n1, '_', ?newPropertyName) AS ?n2).
        BIND (uri(?n2) AS ?newNet).
    """
}

rule_set['create-composite-property-net-from-property-2'] = {
    'label': "create-composite-class-net-from-property-2",
    'comment': ("Create Composite Property Net from relations: ", 
                " (1) between two core properties",
                " (2) with manner property on the same leaf of left property"),
    'construction': """
        # -- New Property Net
        ?newNet a net:Composite_Property_Net ;
            net:isCoreRoleLinked true ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:coverArgNode ?property2Leaf ;
            net:hasPropertyType ?property1Type ;
            net:hasPropertyName ?newPropertyName ;
            net:hasMotherClassNet ?property1Net.
            #net:hasRestriction ?restrictionNet.
            
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify cover(p1:Property, leaf1) and cover(m:Manner, leaf1)
        ?property1Net a [rdfs:subClassOf* net:Property_Net] ; # property 1
            net:isCoreRoleLinked true ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:hasPropertyType ?property1Type ;
            net:hasPropertyName ?property1Name.   
        ?mannerPropertyNet a [rdfs:subClassOf* net:Property_Net] ; # manner property
            net:isCoreRoleLinked true ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:hasPropertyType ?mannerPropertyType ;
            net:hasPropertyName ?mannerPropertyName.
            
        # -- Identify :manner(p2:Property)
        ?mannerPropertyNet amr:role_manner ?property2Net.
        ?property2Net a [rdfs:subClassOf* net:Property_Net] ;
            net:isCoreRoleLinked true ;
			net:coverBaseNode ?property2Leaf ;
			net:hasPropertyName ?property2Name.
            
        # -- condition: disjoint cover
        FILTER NOT EXISTS {?sameBaseLeaf net:coverNode ?node.
			               ?property2Leaf net:coverNode ?node.}
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
        	# -- New Property Name
        BIND (CONCAT(?property2Name, '-', ?property1Name) AS ?newPropertyName).
        	# -- New Property Net
        BIND (CONCAT(str(net:), 'compositeProperty') AS ?n1).
        BIND (CONCAT(?n1, '_', ?newPropertyName) AS ?n2).
        BIND (uri(?n2) AS ?newNet).
    """
}

 
# ---------------------------------------------
# Net Expansion: Add Restriction to Class Net from Property Net
# ---------------------------------------------

rule_set['add-restriction-to-class-net-from-property-1'] = {
    'label': "add-restriction-to-class-net-from-property-1",
    'comment': ("Create and add a Restriction Net: ", 
                " (1) from relation between a class and an individual, ",
                " (2) with a relation corresponding to Non-Core Role property"),
    'construction': """
        # -- New Restriction Net
        ?newNet a net:Restriction_Net ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:coverTargetNode ?argNodeOfBaseNet, ?nodeOfArgNet ;
            net:hasRestrictionOnProperty ?propertyNet ;
            net:hasRestrictionNetValue ?argNet.
            
        # -- New restriction relation
        ?baseNet net:hasRestriction ?newNet.
        
        # -- Propagation of relations (from nodes to nets)
        ?inRelationRole a net:Relation.
        ?inNet ?inRelationRole ?newNet.
        ?outRelationRole a net:Relation.
        ?newNet ?outRelationRole ?outNet.
    """,
    'clause': """
        # -- Identify property(Class, Individual)
        ?propertyNet a [rdfs:subClassOf* net:Property_Net];
            net:isCoreRoleLinked false ;
            net:hasStructure ?sentenceRef ;
            net:coverBaseNode ?baseLeaf ;
            net:hasPropertyRole ?propertyRole ;
            net:hasPropertyName ?propertyName.
        ?baseNet a [rdfs:subClassOf* net:Class_Net] ;
			net:coverBaseNode ?baseLeaf ;
    			net:coverNode ?nodeOfBaseNet ;
			net:hasClassName ?baseName.
        ?argNet a net:Individual_Net ;
			net:coverNode ?nodeOfArgNet ;
            net:hasIndividualLabel ?argLabel.
		?baseNet ?propertyRole ?argNet.
            
        # -- condition: disjoint cover
        FILTER NOT EXISTS {?baseNet net:coverNode ?node.
                           ?argNet net:coverNode ?node.}
            
        # -- Identify inbound relations linked to the base leaf (for propagation) 
        	OPTIONAL {?inNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?inLeaf.
                  ?inLeaf ?inRelationEdge ?baseLeaf.    
                  ?inRelationEdge amr:hasAmrRole ?inRelationRole.}
            
        # -- Identify outgoing relations linked to the base leaf (for propagation) 
        	OPTIONAL {?outNet a [rdfs:subClassOf* net:Net] ;
                       net:coverBaseNode ?outLeaf.
                  ?baseLeaf ?outRelationEdge ?outLeaf.    
                  ?outRelationEdge amr:hasAmrRole ?outRelationRole.}
    """,
    'binding': """
        # -- New Restriction Net
		BIND (CONCAT(str(net:), 'restriction') AS ?rn1).
		BIND (CONCAT(?rn1, '_', str(?propertyName)) AS ?rn2).
		BIND (CONCAT(?rn2, '_', str(?argLabel)) AS ?rn3).
        BIND (uri(?rn3) AS ?newNet).
    """
}