#!/usr/bin/python3.10
# -*-coding:Utf-8 -*

#==============================================================================
# TENET: Net Query Builder
#------------------------------------------------------------------------------
# Class to generate SPARQL query parts related to semantic nets 
#==============================================================================

#==============================================================================
# Importing required modules
#==============================================================================

# --


#==============================================================================
# Data Repository
#==============================================================================

# -- Useful Constant(s)
DEFAULT_ATTRIBUTE_VALUE = f'\"NA\"'
INDENT_STR = '            '
SENTENCE_REF = '?sentenceRef'


#==============================================================================
# Net Class
#==============================================================================

class Net:
    """ Class to generate SPARQL query parts related to semantic nets.
    """
    
    #--------------------------------------------------------------------------
    # Class Attributes
    #--------------------------------------------------------------------------

    # -- Operation Reference 
    INITIALIZED = f'net:initialized'
    RELATION_PROPAGATED = f'net:relation_propagated'
    RESTRICTION_ADDED = f'net:restriction_added'
    FINALIZED = f'net:finalized'
    
    #--------------------------------------------------------------------------
    # Constructor
    #--------------------------------------------------------------------------
              
    def __init__(self, num='', signature=False):
        
        # -- Net Signature
        if not signature:
            self.type_name = 'default'
            self.type_id = 'Net'
            self.id = f'?{self.type_name}Net{num}'
            self.type_uri = f'net:{self.type_id}'
        
        
        # -- Net Attributes
        self.node = f'{self.id}Node'
        self.base_node = f'{self.id}BaseNode'
        self.structure = f'{self.id}Structure'
             
        self.predicate_table = { 
            # *** [attribute_reference: attribute_predicate] ***
            'node': 'coverNode',
            'base_node': 'coverBaseNode',
            'structure': 'hasStructure'
            }
        
        
        # -- Tracker attributes
        self._progress_step = f'{self.id}ProgressStep'
        self._main_net_composante = f'{self.id}MainNetComposante'
        self._net_composante = f'{self.id}NetComposante'
        
        self._track_predicate_table = { 
            # *** [attribute_reference: attribute_predicate] ***
            'progress_step': 'trackProgress',
            'main_net_composante': 'trackMainNetComposante',
            'net_composante': 'trackNetComposante'
            }
        
        
        # -- Private attributes (for relation propagation)
        self._in_relation_role = f'{self.id}InRelationRole'
        self._in_net = f'{self.id}InNet'
        self._out_relation_role = f'{self.id}OutRelationRole'
        self._out_net = f'{self.id}OutNet'

    
    #--------------------------------------------------------------------------
    # Private data accessor(s)
    #--------------------------------------------------------------------------
        
    def __get_predicate(self, attribute_reference):
        predicate_reference = self.predicate_table[f'{attribute_reference}']
        return f'net:{predicate_reference}'
            
    def __get_track_predicate(self, attribute_reference):
        predicate_reference = self._track_predicate_table[f'{attribute_reference}']
        return f'net:{predicate_reference}'

    
    #--------------------------------------------------------------------------
    # Method(s) to track the construction progress
    #--------------------------------------------------------------------------

    def __track_construction_progress(self, progress_step):
        predicate = self.__get_track_predicate('progress_step')
        return f"{self.id} {predicate} {progress_step}."
    
    def get_progress_step(self, progress_step_ref):
        predicate = self.__get_track_predicate('progress_step')
        return f"""
            {{
                # -- Progress Step of {self.id}
                SELECT {self.id} (MAX(?progressStep) AS {progress_step_ref})
            		WHERE {{
                        {self.id} {predicate} ?progressStep.
                }}
                GROUP BY {self.id}
            }}""" 
     
    def is_progress_step(self, progress_step):
        query = self.get_progress_step('?currentProgressStep')
        query += f"""
            FILTER (?currentProgressStep = {progress_step})."""
        return query
    
    def filter_already_processed_net(self, operation_ref):
        predicate = self.__get_track_predicate('progress_step')
        return f"""
            FILTER NOT EXISTS {{ {self.id} {predicate} {operation_ref}. }}"""
 
    def in_progress(self):
        query = self.get_progress_step('?currentProgressStep')
        query += f"""
            FILTER (?currentProgressStep < {Net.FINALIZED})."""
        return query
    

    def __track_net_composante(self, composante, main=False):
        query_code = ''
        predicate = self.__get_track_predicate('net_composante')
        query_code += f"{self.id} {predicate} {composante}."
        if main:
            predicate = self.__get_track_predicate('main_net_composante')
            query_code += f"\n{INDENT_STR}"
            query_code += f"{self.id} {predicate} {composante}."
        query_code += f"\n{INDENT_STR}"
        return query_code
    
    def identify_main_composante(self, target_id):
        predicate = self.__get_track_predicate('main_net_composante')
        return f"\n{INDENT_STR}{self.id} {predicate} {composante}."
    
    def identify_composante(self, target_id):
        predicate = self.__get_track_predicate('net_composante')
        return f"\n{INDENT_STR}{self.id} {predicate} {composante}."
        
    
    #--------------------------------------------------------------------------
    # Method(s) to build 'Construct' parts  
    #--------------------------------------------------------------------------

    def __define_attribute_triples(self, **net_attribute):
        
        query_code = ""
        
        # -- construct triples
        for attr_ref in self.predicate_table.keys():
            if attr_ref in net_attribute.keys():
                predicate = self.__get_predicate(attr_ref)
                attr_value = net_attribute.get(attr_ref)            
                query_code += f"{self.id} {predicate} {attr_value}." 
                query_code += f"\n{INDENT_STR}"
        
        return query_code


    def define_new_net(self, **net_attribute):
        net_attribute.update({'structure': '?sentenceRef'})
        return f"""
            # -- New Net
            {self.id} a {self.type_uri}.
            {self.__define_attribute_triples(**net_attribute)}"""
    
    
    def propagate_relations(self):
        query_code = f"""  
            # -- Propagation of relations (from nodes to nets)
            {self._in_relation_role} a net:Relation.
            {self._in_net} {self._in_relation_role} {self.id}.
            {self._out_relation_role} a net:Relation.
            {self.id} {self._out_relation_role} {self._out_net}."""
        query_code += f"\n{INDENT_STR}"
        query_code += self.__track_construction_progress(Net.RELATION_PROPAGATED) 
        return query_code


    def construct(self, **net_attribute):
        query_code = self.define_new_net(**net_attribute)
        query_code += self.__track_construction_progress(Net.INITIALIZED)
        # -- old --- query_code += self.propagate_relations()
        return query_code
    
    
    def add_attribute(self, **net_attribute):
        query_code = self.__define_attribute_triples(**net_attribute)
        return f"""
            # -- Additional triple(s) for {self.id}
            {query_code}""" 
        
        # for attr_ref in self.predicate_table.keys():
        #     if attr_ref in net_attribute.keys():
        #         predicate = self.__get_predicate(attr_ref)
        #         attr_value = net_attribute.get(attr_ref)     
        #         query_code += f"{self.id} {predicate} {attr_value}." 
        #         query_code += f"\n{INDENT_STR}"  
     
    
    def define_structure(self):
        predicate = self.__get_predicate('structure')  
        query_code = f"{self.id} {predicate} {SENTENCE_REF}." 
        query_code += f"\n{INDENT_STR}"
        
        return f"""
            # -- Additional triple(s) for {self.id}
            {query_code}"""  
            
    def compose(self, *nets):
        query_code = f"\n{INDENT_STR}"
        query_code += f'# -- Composition Tracking for {self.id} \n{INDENT_STR}'
        
        main = True
        for net in nets:
            net_id = f'{net.id}'
            net_node = f'{net.node}'
            query_code += self.__track_net_composante(net_id, main)
            query_code += self.__define_attribute_triples(node=net_node)
            main = False
           
        if (nets is not None) & (len(nets) != 0):
            query_code += f'\n{INDENT_STR}'
            query_code += f'# -- Deprecate net \n{INDENT_STR}'
            query_code += f'{nets[0].id} a net:Deprecated_Net.\n{INDENT_STR}'
            
        return query_code
            
    def deprecate(self):
        return f"""
            # -- Deprecate net
            """
    
    
    #--------------------------------------------------------------------------
    # Method(s) to build 'Complement Clause' parts  
    #--------------------------------------------------------------------------
    
    def identify_structure(self):
        return f"""
            # -- Identify structure
            ?root a amr:AMR_Root.
            ?root amr:hasSentenceID {SENTENCE_REF}.
        """
    
    def identify_relations_for_propagation(self, base_node=''):
        if base_node=='':
            base_node = self.base_node
        return f"""
            # -- Identify inbound relations linked to the base leaf (for propagation)
            OPTIONAL {{ 
                          {self._in_net} a [rdfs:subClassOf* net:Net] ;
                              net:coverBaseNode ?inLeaf.
                          ?inLeaf ?inRelationEdge {base_node}.    
                          ?inRelationEdge amr:hasAmrRole {self._in_relation_role}.
                      }}
            
            # -- Identify outgoing relations linked to the base leaf (for propagation)
            OPTIONAL {{ 
                        {self._out_net} a [rdfs:subClassOf* net:Net] ;
                            net:coverBaseNode ?outLeaf.
                        {base_node} ?outRelationEdge ?outLeaf.    
                        ?outRelationEdge amr:hasAmrRole {self._out_relation_role}.
                      }}
        """
    
    # def identify_relations_for_propagation(self, base_node):
    #     return f"""
    #         # -- Identify inbound relations linked to the base leaf (for propagation)
    #         OPTIONAL {{ {{
    #             SELECT {self._in_net} {self._in_relation_role}
    #             WHERE {{
    #                       {self._in_net} a [rdfs:subClassOf* net:Net] ;
    #                           net:coverBaseNode ?inLeaf.
    #                       ?inLeaf ?inRelationEdge {base_node}.    
    #                       ?inRelationEdge amr:hasAmrRole {self._in_relation_role}.
    #                   }}
    #                  }} }}
            
    #         # -- Identify outgoing relations linked to the base leaf (for propagation)
    #         OPTIONAL {{ {{
    #             SELECT {self._out_net} {self._out_relation_role}
    #             WHERE {{
    #                     {self._out_net} a [rdfs:subClassOf* net:Net] ;
    #                         net:coverBaseNode ?outLeaf.
    #                     {base_node} ?outRelationEdge ?outLeaf.    
    #                     ?outRelationEdge amr:hasAmrRole {self._out_relation_role}.
    #                   }}
    #                  }} }}
    #     """

    def complete_clauses_for_construction(self, base_node=''):
        query_code = self.identify_structure()
        query_code += self.identify_relations_for_propagation(base_node)
        return query_code
    
    
    #--------------------------------------------------------------------------
    # Method(s) to build 'Identification Clause' parts  
    #--------------------------------------------------------------------------

    def __select_all_attribute_triples(self, **net_attribute):
        
        query_code = ""
        
        # -- construct triples with default object for non-declared attributes
        for attr_ref in self.predicate_table.keys():
            
            predicate = self.__get_predicate(attr_ref)
            
            if attr_ref in net_attribute.keys():
                attr_value = net_attribute.get(attr_ref)
                optional_start = ''
                optional_end = ''
                comment = ''
            else:
                attr_value = self.__dict__[attr_ref]
                optional_start = 'OPTIONAL {{ '
                optional_end = ' }}'
                comment = ''
                
            query_code += f"{optional_start}"
            query_code += f"{self.id} {predicate} {attr_value}.{comment}" 
            query_code += f"{optional_end}"             
            query_code += f"\n{INDENT_STR}"
        
        return query_code
    
    
    def __select_targeted_attribute_triples(self, **net_attribute):
        
        query_code = ""
        
        # -- select triples
        for attr_ref in self.predicate_table.keys():
            
            predicate = self.__get_predicate(attr_ref)
            
            if attr_ref in net_attribute.keys():
                attr_value = net_attribute.get(attr_ref)
                comment = ''
                
                query_code += f"{self.id} {predicate} {attr_value}.{comment}"          
                query_code += f"\n{INDENT_STR}"
        
        return query_code
    
    
    def __select_optional_attribute_triples(self, **net_attribute):
        
        query_code = ""
        
        # -- construct triples with default object for non-declared attributes
        for attr_ref in self.predicate_table.keys():
            
            predicate = self.__get_predicate(attr_ref)
            
            if attr_ref not in net_attribute.keys():
                attr_value = self.__dict__[attr_ref]
                optional_start = 'OPTIONAL {{ '
                optional_end = ' }}'
                comment = ''
                
                query_code += f"{optional_start}"
                query_code += f"{self.id} {predicate} {attr_value}.{comment}" 
                query_code += f"{optional_end}"             
                query_code += f"\n{INDENT_STR}"
        
        return query_code
               
  
    def __select_node(self):
        
        query_code = ""
        
        optional_start = 'OPTIONAL {{ '
        optional_end = ' }}'
        select_predicates = f"""(net:coverNode|net:coverBaseNode|net:coverTargetNode|net:coverArgNode)"""
        
        query_code += f"{optional_start}"
        query_code += f"# -- Select nodes of net {self.id} \n{INDENT_STR}" 
        query_code += f"{self.id} {select_predicates} {self.node}." 
        query_code += f"{optional_end}"  
    
        return query_code
            
            
    def identify_net(self, **net_attribute):
        return f"""
            # -- Identify net
            {self.id} a [rdfs:subClassOf* {self.type_uri}].
            FILTER NOT EXISTS {{ {self.id} a net:Deprecated_Net. }}
            {self.__select_all_attribute_triples(**net_attribute)}
            {self.__select_node()}"""  
                
            
    def identify_net_without_optional_attribute(self, **net_attribute):
        return f"""
            # -- Identify net
            {self.id} a [rdfs:subClassOf* {self.type_uri}].
            FILTER NOT EXISTS {{ {self.id} a net:Deprecated_Net. }}
            {self.__select_targeted_attribute_triples(**net_attribute)}
            {self.__select_node()}"""  
            
    
    def identify_relation(self, relation_role, target_id):
        return f"""
            # -- Identify relation of {self.id}
            {self.id} {relation_role} {target_id}."""  
  
    
    def identify_composante(self, target_id):
        predicate = self.__get_track_predicate('net_composante')
        return f"""
            # -- Identify net composante of {self.id}
            {self.id} {predicate} {target_id}."""  
  
    
    def identify_main_composante(self, target_id):
        predicate = self.__get_track_predicate('main_net_composante')
        return f"""
            # -- Identify main net composante of {self.id}
            {self.id} {predicate} {target_id}."""  
    
    
    #--------------------------------------------------------------------------
    # Method(s) to build 'Binding' parts  
    #--------------------------------------------------------------------------
    
    def bind_uri(self, net_name='nameless', node_reference='blank'):
        
        ref1 = f"{self.id}Ref1"
        ref2 = f"{self.id}Ref2"
        ref3 = f"{self.id}Ref3"
        
        if net_name=='nameless':
            net_name = f"'nameless'"
            refNet = f"?namelessRefNet"
        else:
            refNet = f"{net_name}RefNet"
            
        if node_reference=='blank':
            node_reference = f"'blankNode'"
            refNode = f"?blankRefNode"
        else:
            refNode = f"{node_reference}RefNode"
            
        return f"""
            # -- New Net
            BIND (REPLACE({net_name}, ' ', "") AS {refNet}).
            BIND (REPLACE({node_reference}, ' ', "") AS {refNode}).
            BIND (CONCAT(str(net:), '{self.type_name}') AS {ref1}).
            BIND (CONCAT({ref1}, '_', {refNet}) AS {ref2}).
            BIND (CONCAT({ref2}, '_', {refNode}) AS {ref3}).
            BIND (uri({ref3}) AS {self.id})."""